/*
 * $Id: ng_grep.c,v 1.33 2014/04/30 11:01:05 andy Exp $
 *
 * Netgraph "grep" node
 *
 * full-text search for data field ot tcp packets 
 * on the basis of ng_echo.c from FreeBSD 10.0
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/syslog.h>
#include <netgraph/ng_message.h>
#include <netgraph/netgraph.h>
#include <netgraph/ng_parse.h>
#include "ng_grep.h"
#include "my_pattern.h"

/* Netgraph methods */
static ng_constructor_t	ng_grep_cons;
static ng_shutdown_t	ng_grep_shut;
static ng_newhook_t	ng_grep_newhook;
static ng_rcvmsg_t	ng_grep_rcvmsg;
static ng_rcvdata_t	ng_grep_rcvdata;
static ng_disconnect_t	ng_grep_discon;

/* Parsing for set/get messages */
// -- getstats (clearstats, getclearstats too)
static const struct ng_parse_struct_field ng_grep_stats_out_type_fields[] = NG_GREP_STATS_OUT_TYPE_INFO;

static const struct ng_parse_type ng_grep_stats_out_type = {
	&ng_parse_struct_type,
	&ng_grep_stats_out_type_fields
};

// --add (and del, and check too)
static const struct ng_parse_struct_field ng_grep_add_in_type_fields[] = NG_GREP_ADD_IN_TYPE_INFO;

static const struct ng_parse_type ng_grep_add_in_type = {
	&ng_parse_struct_type,
	&ng_grep_add_in_type_fields
};

static const struct ng_parse_struct_field ng_grep_add_out_type_fields[] = NG_GREP_ADD_OUT_TYPE_INFO;

static const struct ng_parse_type ng_grep_add_out_type = {
	&ng_parse_struct_type,
	&ng_grep_add_out_type_fields
};

// --flush
static const struct ng_parse_struct_field ng_grep_flush_out_type_fields[] = NG_GREP_FLUSH_OUT_TYPE_INFO;

static const struct ng_parse_type ng_grep_flush_out_type = {
	&ng_parse_struct_type,
	&ng_grep_flush_out_type_fields
};

// -- summary cmdlist
static const struct ng_cmdlist ng_grep_cmdlist[] = {
	{
		NG_GREP_COOKIE,
		NGM_GREP_GETSTATS,
		"getstats",
		NULL,
		&ng_grep_stats_out_type
	},
	{
		NG_GREP_COOKIE,
		NGM_GREP_CLEARSTATS,
		"clearstats",
		NULL,
		NULL,
	},
	{
		NG_GREP_COOKIE,
		NGM_GREP_GETCLEARSTATS,
		"getclearstats",
		NULL,
		&ng_grep_stats_out_type
	},
	{
		NG_GREP_COOKIE,
		NGM_GREP_ADD,
		"add",
		&ng_grep_add_in_type,
		&ng_grep_add_out_type
	},
	{
		NG_GREP_COOKIE,
		NGM_GREP_DEL,
		"del",
		&ng_grep_add_in_type,
		&ng_grep_add_out_type
	},
	{
		NG_GREP_COOKIE,
		NGM_GREP_CHECK,
		"check",
		&ng_grep_add_in_type,
		&ng_grep_add_out_type
	},
	{
		NG_GREP_COOKIE,
		NGM_GREP_FLUSH,
		"flush",
		NULL,
		&ng_grep_flush_out_type
	},
	{0}
};

/* Netgraph type */
static struct ng_type typestruct = {
	.version =	NG_ABI_VERSION,
	.name =		NG_GREP_NODE_TYPE,
	.constructor =	ng_grep_cons,
	.shutdown =	ng_grep_shut,
	.newhook =	ng_grep_newhook,
	.rcvmsg =	ng_grep_rcvmsg,
	.rcvdata =	ng_grep_rcvdata,
	.disconnect =	ng_grep_discon,
	.cmdlist =	ng_grep_cmdlist,
};
NETGRAPH_INIT(grep, &typestruct);

/* Node private data */
struct ng_grep_private {
	pat_p	patterns;		/* pattern chain */
        hook_p 	in;
        hook_p 	meet;
        hook_p 	miss;
	u_long	in_c;
	u_long	bad_c;
	u_long	meet_c;
	u_long	miss_c;
        node_p 	node;                   /* Our netgraph node */
};
typedef struct ng_grep_private *priv_p;


/* Constructor */
static int ng_grep_cons(node_p node){
	priv_p	priv;

//	log(LOG_DEBUG,"ng_grep: ------------------------ construct!\n");

	priv = malloc(sizeof(*priv), M_NETGRAPH_GREP, M_ZERO | M_WAITOK);
	NG_NODE_SET_PRIVATE(node, priv);
	priv->node = node;
	priv->patterns = patterns_chain_init();
	priv->in_c = 0;
	priv->bad_c = 0;
	priv->meet_c = 0;
	priv->miss_c = 0;
	return (0);
}

static int ng_grep_newhook(node_p node, hook_p hook, const char *name){

	priv_p	priv = NG_NODE_PRIVATE(node);
	hook_p	*localhook;

	if (strcmp(name, NG_GREP_HOOK_IN) == 0) { localhook = &priv->in;
        } else if (strcmp(name, NG_GREP_HOOK_MEET) == 0) { localhook = &priv->meet;
        } else if (strcmp(name, NG_GREP_HOOK_MISS) == 0) { localhook = &priv->miss;
        } else return (EINVAL);

        if (*localhook != NULL) return (EISCONN);
        *localhook = hook;
        NG_HOOK_SET_PRIVATE(hook, localhook);

	return(0);
}

/* Receive control message. We just bounce it back as a reply. */
static int ng_grep_rcvmsg(node_p node, item_p item, hook_p lasthook){
	struct ng_mesg	*msg;
	struct ng_mesg	*resp 	= NULL;
	priv_p 		priv 	= NG_NODE_PRIVATE(node);
	int 		error	= 0;
	

//	log(LOG_DEBUG,"ng_grep: ------------------------ rcvmsg!\n");

	NGI_GET_MSG(item, msg);

	switch(msg->header.typecookie){
	case NG_GREP_COOKIE:
		switch(msg->header.cmd){
			case NGM_GREP_GETSTATS:
			case NGM_GREP_CLEARSTATS:
			case NGM_GREP_GETCLEARSTATS:{
				struct	ng_grep_stats_out *stats;

				NG_MKRESPONSE(resp, msg, sizeof(*stats), M_NOWAIT);
				if(!resp){
					error = ENOMEM;
					break;
				}
				stats = (struct ng_grep_stats_out *)resp->data;
				if(msg->header.cmd != NGM_GREP_CLEARSTATS){
					stats->in_c = priv->in_c;
					stats->bad_c = priv->bad_c;
					stats->miss_c = priv->miss_c;
					stats->meet_c = priv->meet_c;
				}
				if(msg->header.cmd != NGM_GREP_GETSTATS){
					priv->in_c = 0;
					priv->bad_c = 0;
					priv->meet_c = 0;
					priv->miss_c = 0;
				}
				break;
			}
			case NGM_GREP_FLUSH:{
				struct	ng_grep_flush_out *result;

				NG_MKRESPONSE(resp, msg, sizeof(*result), M_NOWAIT);
				if(!resp){
					error = ENOMEM;
					break;
				}
				result = (struct ng_grep_flush_out *)resp->data;
				result->rc=patterns_chain_flush(priv->patterns);
				break;
			}
			case NGM_GREP_ADD:
			case NGM_GREP_DEL:
			case NGM_GREP_CHECK:{
//				void		**prec	= NULL;
				pat_p		found	= NULL;
				struct	ng_grep_add_in 	*income;
				struct	ng_grep_add_out *result;
				u_char	*str_in;
				u_char	pat_flags;
				
				income = (struct ng_grep_add_in *)msg->data;
	
//				log(LOG_DEBUG,"ng_grep: ------------------------ rcvmsg::add '%s'\n",pattern->raw);

				NG_MKRESPONSE(resp, msg, sizeof(*result), M_NOWAIT);
				if(!resp){
					error = ENOMEM;
					break;
				}
				result = (struct ng_grep_add_out *)resp->data;

				if(strlen(income->url)){
//					log(LOG_DEBUG,"ng_grep::rcvmsg::add it's  URL\n");
					str_in = income->url;
					pat_flags = P_FLAG_URL;
					strcpy(result->url, str_in);
				}else{
//					log(LOG_DEBUG,"ng_grep::rcvmsg::add it's not URL, assuming RAW\n");
					str_in = income->raw;
					pat_flags = P_FLAG_I_C;
					strcpy(result->raw, str_in);
				}

				if(income->flag_i_c){pat_flags |= P_FLAG_I_C;}
				if(income->flag_rex){pat_flags |= P_FLAG_REX;}
				if(income->flag_sub){pat_flags |= P_FLAG_SUB;}

				switch(msg->header.cmd){
					case NGM_GREP_ADD:{
						result->rc = pattern_add(priv->patterns, str_in, (void**)&found, pat_flags);
						break;
					}
					case NGM_GREP_DEL:{
						result->rc = pattern_del(priv->patterns, str_in, (void**)&found, pat_flags);
						break;
					}
					case NGM_GREP_CHECK:{
						result->rc = pattern_check(priv->patterns, str_in, (void**)&found, pat_flags);
						break;
					}
				}
//				log(LOG_DEBUG,"ng_grep: ----- rcvmsg::add_url ----- check 1 after flight\n");
				//if((pat_flags & P_FLAG_URL) && prec){
//				result->flag_i_c = pat_flags & P_FLAG_I_C;
//				result->flag_rex = pat_flags & P_FLAG_REX;
//				result->flag_sub = pat_flags & P_FLAG_SUB;
//				if(pat_flags & P_FLAG_I_C){result->flag_i_c = 1;}
//				if(pat_flags & P_FLAG_REX){result->flag_rex = 1;}
//				if(pat_flags & P_FLAG_SUB){result->flag_sub = 1;}
				if(found){
					if(found->precomp){
//						log(LOG_DEBUG,"ng_grep: ----- rcvmsg::add_url ----- check 2 after flight\n");
						strncpy(result->host,((p_url_p)found->precomp)->host, ((p_url_p)found->precomp)->host_l);
						strncpy(result->path,((p_url_p)found->precomp)->path, ((p_url_p)found->precomp)->path_l);
					}
					result->flag_i_c = found->flag_i_c;
					result->flag_sub = found->flag_sub;
					result->flag_rex = found->flag_rex;
					result->meet_c = found->meet_c;
				}
				break;
			}
			default:
				error = EINVAL;		/* unknown command */
				break;
		}
	default:
		error = EINVAL;				/* not for me */
		break;
	}

	NG_RESPOND_MSG(error, node, item, resp);
	NG_FREE_MSG(msg);
	return (error);
}

/*
 * Receive data
 */
static int ng_grep_rcvdata(hook_p hook, item_p item){

	const priv_p    priv 	= NG_NODE_PRIVATE(NG_HOOK_NODE(hook));
	int 		error	= 0;
	struct mbuf 	*m;
	uint8_t		*p;
	u_int		data_len;
	u_int		data_beg;

	priv->in_c++;

	if(hook==priv->miss || hook==priv->meet){
		log(LOG_ERR,"ng_grep::rcvdata from out-only hook, discard!\n");
		NG_FREE_ITEM(item);
		error=EINVAL;
		priv->bad_c++;
		return (error);
	}

	m=item->body.da_m;
	p=mtod(m,uint8_t *);
// Ahtung!! Byte-order is critical here!! ----------------------------------------------------------------------
// I know it's bad. Will be corrected asap. I hope. Probably.
	if(((p[0] & 0xf0) == 0x40) && (p[9] == 0x06)){		// is ipv4? is tcp ?
		data_beg = (p[0] & 0x0f) * 4;			// ip hdr length
		data_len = (p[2] << 8) + p[3] - 2;		// pkt length 2 is magic
		data_beg += (p[32] & 0xf0) >> 2;		// tcp hdr length
// End of byte-order critical code -----------------------------------------------------------------------------
		if(data_len > data_beg){
			data_len -= data_beg;
//			log(LOG_DEBUG,"ng_grep::rcvdata data start '%d', data length '%d' \n",data_beg,data_len);
			error = check_incom_string(priv->patterns, &p[data_beg], data_len);
		}else{
			error = ENOSYS;
		}
	}else{
		error = ENOSYS;
	}

//	error = 1;						// DBG ONLY!! Remember to remove!

	if(error){
		NG_FWD_ITEM_HOOK(error, item, priv->miss);
		priv->miss_c++;
	}else{
		NG_FWD_ITEM_HOOK(error, item, priv->meet);
		priv->meet_c++;
	}
	return (error);
}

/*
 * Removal of the last link destroys the nodeo
 */
static int ng_grep_discon(hook_p hook){
//	log(LOG_DEBUG,"ng_grep: ------------------------ disconnect!\n");
	if ((NG_NODE_NUMHOOKS(NG_HOOK_NODE(hook)) == 0)
	&& (NG_NODE_IS_VALID(NG_HOOK_NODE(hook)))) {
		ng_rmnode_self(NG_HOOK_NODE(hook));
	}
	return (0);
}

static int ng_grep_shut(node_p node) {
        const priv_p    priv = NG_NODE_PRIVATE(node);
//	log(LOG_DEBUG,"ng_grep: ------------------------ shutdown!\n");

        NG_NODE_SET_PRIVATE(node, NULL);
        NG_NODE_UNREF(node);
//        free(priv->patterns, M_NETGRAPH_GREP);
        patterns_chain_free(priv->patterns);
        free(priv, M_NETGRAPH_GREP);

        return (0);
}
